/*
 * launcherMainWindow.cpp
 *
 *  Created on: 10 sept. 2012
 *      Author: valot
 */

#include "launcherMainWindow.h"
#include "../lib/consoleout.h"
#include <QtGui>
#include <QDockWidget>
#include <QCloseEvent>
#include <QMessageBox>
#include <QMenuBar>
#include <QAction>
#include <QGroupBox>
#include <QLabel>
#include <QApplication>
#include <QDir>
#include <QFileDialog>
#include <QCheckBox>
#include "../lib/share/utilities.h"
#include "logQIODevice.h"


#include "ui_mcqgui.h"


LauncherMainWindow::LauncherMainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::GuiMainWindow)
{
    ui->setupUi(this);
    _number_of_cpus = 1;
    QFileInfo mcq_icon;
    mcq_icon.setFile(MASSCHROQ_ICON);
    if (mcq_icon.exists()) {
        setWindowIcon(QIcon(mcq_icon.filePath()));
    } else {
        QString mcq_icon_current(QApplication::applicationDirPath());
        mcq_icon_current.append("/").append(mcq_icon.fileName());
        mcq_icon.setFile(mcq_icon_current);
        if (mcq_icon.exists()) {
            setWindowIcon(QIcon(mcq_icon.filePath()));
        }
    }

    _masschroqRunningThread = 0;
    //_logQTextEdit=0;
    _masschroq = new MassChroq();
    _masschroq->setTmpDir(QDir::tempPath());
    this->initialiseMassChroqRunningThread();
    this->addMenu();
    this->addWidget();

    setAttribute(Qt::WA_AlwaysShowToolTips, true);
}

LauncherMainWindow::~LauncherMainWindow() {
    delete (_masschroqRunningThread);
    delete ui;
}

void LauncherMainWindow::closeEvent(QCloseEvent * event) {
    if (maybeSave()) {
        _masschroqRunningThread->canceled();
        _masschroqRunningThread->terminate();
        event->accept();
    } else {
        event->ignore();
    }
}

bool LauncherMainWindow::maybeSave() {
    QMessageBox::StandardButton ret;
    ret =
        QMessageBox::warning(this, tr("MassChroQ GUI"),
                             tr(
                                 "Do you want to quit MassChroQ Gui and stop all current processing?"),
                             QMessageBox::Ok | QMessageBox::Cancel);
    if (ret == QMessageBox::Ok)
        return true;
    else if (ret == QMessageBox::Cancel)
        return false;
    else
        return true;
}

void LauncherMainWindow::initialiseMassChroqRunningThread() {
    qDebug() << "initialiseMassChroqRunningThread";
    if (_masschroqRunningThread != 0) {
        delete (_masschroqRunningThread);
    }
    _masschroqRunningThread = new MasschroqRunningThread();
    //add connection
    _masschroqRunningThread->setMaxProgress(4);
    connect(_masschroqRunningThread, SIGNAL(finishRunning()), this,
            SLOT(finishRunningMassChroq()));
    connect(_masschroqRunningThread, SIGNAL(errorDuringRunning(QString)), this,
            SLOT(viewErrorDuringRunning(QString)));
    qDebug() << "Finish initialiseMassChroqRunningThread";
}

void LauncherMainWindow::addMenu() {
    qDebug() << "LoadFileMenus begin";


    ui->actionQuit->setShortcuts(QKeySequence::Quit);
    connect(ui->actionQuit, SIGNAL(triggered()), this, SLOT(close()));


    connect(ui->actionAbout, SIGNAL(triggered()), this, SLOT(about()));

    qDebug() << "LoadFileMenus end";
}

void LauncherMainWindow::addWidget() {

    qDebug() << "addLogWidget begin";
    //_logQTextEdit = new LogQTextEdit(this);
    LogQIODevice * logQIODevice = new LogQIODevice();
    ConsoleOut::setCout(new QTextStream(logQIODevice));
    ConsoleOut::setCerr(new QTextStream(logQIODevice));
    //setCentralWidget(_logQTextEdit);
    connect(logQIODevice, SIGNAL(appendLogString(QString)), this, SLOT(appendLogString(QString)));

    mcqout() << "Initialising logger" << endl;

    this->addMasschroqmlSelectionGroup();

    ui->parseOnlyBox->setChecked(false);

    //QCheckBox * only = new QCheckBox(tr("Only parse peptide file"));

    //only->setChecked(_onlyPeptideParse);
    //only->setToolTip(tr("If the identified peptides are given to masschroq via peptide text files (defined in FILE),\n"
    //		"He parses them, creates a new file called parsed-peptides_FILE which contains the FILE content plus the parsed peptides information."));
    //connect(ui->parseOnlyBox, SIGNAL(clicked(bool)), this, SLOT(modifyPeptideParse(bool)));
    //_mainlayout->addWidget(only);

    this->addTempDirectorySelectionGroup();
    this->addCpuNumberSelectionGroup();


    //loadButton = new QPushButton(tr("&Start MassChroQ"));
    ui->startPushButton->setDefault(true);
    connect(ui->startPushButton, SIGNAL(clicked()), this, SLOT(startMassChroq()));
}

void LauncherMainWindow::addMasschroqmlSelectionGroup() {

    ui->filePathLineEdit->setText(tr("No file Selected"));
    ui->filePathLineEdit->setReadOnly(true);
    ui->fileSelectPushButton->setDefault(true);
    connect(ui->fileSelectPushButton, SIGNAL(clicked()), this,
            SLOT(selectMassChroqFile()));
}

void LauncherMainWindow::addTempDirectorySelectionGroup() {
    ui->tempDirLineEdit->setText(_masschroq->getTmpDirName());
    ui->tempDirLineEdit->setReadOnly(true);
    ui->tempDirPushButton->setDefault(true);
    connect(ui->tempDirPushButton, SIGNAL(clicked()), this,
            SLOT(selectTempDirectory()));
}

void LauncherMainWindow::addCpuNumberSelectionGroup() {

    //ui->cpuNumberSpinBox->se
    //cpu_number->setDecimals(0);
    int i = QThread::idealThreadCount();
    if (i == -1 ) i = 1;
    ui->cpuNumberSpinBox->setMaximum(i);
    ui->cpuNumberSpinBox->setMinimum(1);
    ui->cpuNumberSpinBox->setSingleStep(1);
    ui->cpuNumberSpinBox->setValue(1);


    connect(ui->cpuNumberSpinBox, SIGNAL(valueChanged(int)), this,
            SLOT(setNumberOfCpus(int)));


}

void LauncherMainWindow::setNumberOfCpus(int z) {
    _number_of_cpus = z;
}

void LauncherMainWindow::selectMassChroqFile() {
    QString filename = QFileDialog::getOpenFileName(this,
                       tr("Open MassChroqML File"), QString::null,
                       tr("MassChroqML files (*.masschroqML *.xml)"));

    if (filename.isEmpty()) {
        return;
    }
    try {
        _masschroq->setXmlFilename(filename);
        QFileInfo filenameInfo(filename);
        ui->filePathLineEdit->setText(filenameInfo.fileName());
    } catch (mcqError& error) {
        viewError(error.qwhat());
    }
}

void LauncherMainWindow::selectTempDirectory() {
    QString fdirname = QFileDialog::getExistingDirectory(this,
                       tr("Open Temporary directory"), QString::null,
                       QFileDialog::ShowDirsOnly);
    if (fdirname.isEmpty()) {
        return;
    }
    try {
        _masschroq->setTmpDir(fdirname);
        ui->tempDirLineEdit->setText(fdirname);
    } catch (mcqError& error) {
        viewError(error.qwhat());
    }
}

void LauncherMainWindow::startMassChroq() {
    qDebug() << "LauncherMainWindow::startMassChroq Begin";
    QString filename = _masschroq->getXmlFilename();
    if (filename.isEmpty()) {
        this->viewError(tr("No MassChroqML file selected"));
        return;
    }
    MassChroq * masschroqTmp = new MassChroq();
    try {
        masschroqTmp->setMasschroqDir(QApplication::applicationDirPath());
        masschroqTmp->setXmlFilename(filename);
        masschroqTmp->setTmpDir(_masschroq->getTmpDirName());
        Quantificator::setCpuNumber(_number_of_cpus);
    } catch (mcqError& error) {
        viewError(error.qwhat());
    }
    //Button pas to stop
    ui->startPushButton->setText(tr("&Stop"));
    disconnect(ui->startPushButton, SIGNAL(clicked()), this, SLOT(startMassChroq()));
    connect(ui->startPushButton, SIGNAL(clicked()), this, SLOT(AbordMassChroqRunning()));
    //TImer
    _dt_begin = QDateTime::currentDateTime();
    //Remove Text on logger
    ui->mcqLogTextEdit->setText("");
    //Start Running
    _masschroqRunningThread->onlyParsePeptide(ui->parseOnlyBox->isChecked());
    _masschroqRunningThread->runMassChroqML(masschroqTmp);
}

void LauncherMainWindow::AbordMassChroqRunning() {
    if (this->close()) {
        //Stop Thread
        qDebug() << "AbordMassChroqRunning";
        _masschroqRunningThread->canceled();
        _masschroqRunningThread->terminate();
        qDebug() << "Finish AbordMassChroqRunning";
    }
}
void LauncherMainWindow::finishRunningMassChroq() {
    //Add reinitialisation of button
    ui->startPushButton->setText(tr("&Start MassChroQ"));
    disconnect(ui->startPushButton, SIGNAL(clicked()), this,
               SLOT(AbordMassChroqRunning()));
    connect(ui->startPushButton, SIGNAL(clicked()), this, SLOT(startMassChroq()));
    this->initialiseMassChroqRunningThread();
    //Add Time
    QDateTime dt_end = QDateTime::currentDateTime();

    Duration dur = Utilities::getDurationFromDates(_dt_begin, dt_end);

    mcqout() << "MassChroQ finished"<<endl;
    mcqout() << "MassChroQ execution time : "
             << Utilities::getDaysFromDuration(dur) << " days, "
             << Utilities::getHoursFromDuration(dur) << " hours, "
             << Utilities::getMinutesFromDuration(dur) << " minutes, "
             << Utilities::getSecondsFromDuration(dur) << " seconds."
             << endl<< endl;
}
void LauncherMainWindow::viewErrorDuringRunning(QString error) {
    //Add reinitialisation of button
    ui->startPushButton->setText(tr("&Start MassChroQ"));
    disconnect(ui->startPushButton, SIGNAL(clicked()), this,
               SLOT(AbordMassChroqRunning()));
    connect(ui->startPushButton, SIGNAL(clicked()), this, SLOT(startMassChroq()));
    this->initialiseMassChroqRunningThread();
    //view Error
    this->viewError(error);
}
void LauncherMainWindow::viewError(QString error) {
    QMessageBox::warning(this,
                         tr("Oops! an error occurred in MassChroQ. Dont Panic :"), error);
}

void LauncherMainWindow::about() {
    QMessageBox::about(this, tr("About MassChroQ"),
                       tr("This is <b>MassChroQ Gui</b> version %1").arg(
                           MASSCHROQ_VERSION));
}


void LauncherMainWindow::appendLogString(QString log)
{
    ui->mcqLogTextEdit->moveCursor(QTextCursor::End);
    ui->mcqLogTextEdit->insertPlainText(log);
}
