/**
 * \file quantificator.cpp
 * \date August 03, 2011
 * \author Edlira Nano
 */

#include "quantificator.h"
#include "./consoleout.h"
#include "xic/xic_base.h"
#include "xic/xicfilterdetectmatch.h"
#include "quantifications/mapquanti.h"
#include "quantifications/mapquantirealxic.h"
#include "quanti_items/quantiItemPeptide.h"
#include "quanti_items/quantiItemMzRt.h"
#include "quanti_items/quantiitempeptidenaturalisotope.h"
#include "quantifications/quantificationMethod.h"
#include "peptides/peptide_list.h"
#include <pappsomspp/peptide/peptidenaturalisotopelist.h>

#include <QStringList>
#include <algorithm>
#include <QtConcurrentMap>
#include <QDir>
#include <pappsomspp/xic/filters/xicfilterinterface.h>
#include <pappsomspp/xic/detection/xicdetectioninterface.h>
#include "peptides/peptidert.h"

uint Quantificator::_cpu_number = 1;


void Quantificator::setCpuNumber(uint cpu_number) {
    qDebug() << "Quantificator::setCpuNumber" <<  cpu_number;
    uint ideal_number_of_thread = (uint) QThread::idealThreadCount();
    if (cpu_number > ideal_number_of_thread) {
        cpu_number = ideal_number_of_thread;
    }
    else {
        QThreadPool::globalInstance()->setMaxThreadCount(cpu_number);
    }
    _cpu_number = cpu_number;
    mcqout() << QObject::tr("%1 CPUs used").arg(cpu_number) << endl;
}


Quantificator::Quantificator(const QString & xml_id,
                             const msRunHashGroup * quanti_group,
                             QuantificationMethod * quanti_method) :
    _xml_id(xml_id) {
    set_quanti_group(quanti_group);
    set_quanti_method(quanti_method);
    _quanti_items = new std::vector<QuantiItemBase *>;
}

Quantificator::~Quantificator() {
    std::vector<QuantiItemBase *>::iterator it;
    for (it = _quanti_items->begin(); it != _quanti_items->end(); ++it) {
        if (*it != 0) {
            delete (*it);
            *it = 0;
        }
    }

    if (_quanti_items != 0) {
        delete _quanti_items;
        _quanti_items = 0;
    }

}


void Quantificator::setTracesOutput(QDir traces_directory, McqTsvFormat traces_format) {
    _traces_directory = traces_directory;
    _traces_format = traces_format;
}


void Quantificator::set_quanti_group(const msRunHashGroup * quanti_group) {
    _group_to_quantify = quanti_group;
}

void Quantificator::set_quanti_method(QuantificationMethod * quanti_method) {
    _quanti_method = quanti_method;
}

void Quantificator::setMatchingMode(McqMatchingMode match_mode) {
    _matching_mode = match_mode;
}

const msRunHashGroup *
Quantificator::getMsRunGroup() const {
    return (_group_to_quantify);
}


void Quantificator::computeRetentionTimeReferences() {
    //_map_peptides_2_retention_times.clear();
    //_accumulator.clear();
    for (const Peptide * p_peptide:*_p_peptide_list) {

        PeptideRt * _p_peptide_rt = p_peptide->getPeptideRtSp().get();
        auto find_it = std::find(_peptide_rt_list.begin(),_peptide_rt_list.end(),  _p_peptide_rt);
        if (find_it == _peptide_rt_list.end()) {
            _peptide_rt_list.push_back(_p_peptide_rt);
            // auto it_new =  _map_peptides_2_retention_times.insert(std::pair<const Peptide *, std::map< const Msrun * , std::vector< MsMsRtIntensity > > >(p_peptide,std::map< const Msrun * , std::vector< MsMsRtIntensity > >())).first;
            //auto it_charge_new = map_peptides_charge_set.insert(std::pair<const Peptide *, std::set<unsigned int>>(p_peptide, std::set<unsigned int>())).first;

            for (auto & pair : *_group_to_quantify) {
                const Msrun * p_msrun = pair.second;
                //auto it_rt_list = it_new->second.insert(std::pair<const Msrun * , std::vector< MsMsRtIntensity >>
                //                                        (p_msrun, std::vector< MsMsRtIntensity >())).first;

                //qDebug() << "Quantificator::computeRetentionTimeReferences getObservedInList size" << p_peptide->getObservedInList().size();
                for (auto && obs_in:p_peptide->getObservedInList()) {
                    if(obs_in->getPmsRun() == p_msrun) {
                        MsMsObservation observation(p_msrun);
                        observation.intensity = obs_in->getIntensity();
                        observation.rt = obs_in->getRt();
                        observation.z = obs_in->getZ();
                        observation.mz = p_peptide->getMz(observation.z);
                        //it_rt_list->second.push_back(observation);

                        //it_charge_new->second.insert(observation.z);

                        _p_peptide_rt->addMsMsRtIntensity(observation);
                    }
                }
            }
        }
    }

    if ((_matching_mode == McqMatchingMode::mean) || ( _matching_mode == McqMatchingMode::real_or_mean) || ( _matching_mode == McqMatchingMode::post_matching)) {
        //accumulatorToRetentionTimeReferences();
        for (auto && p_peptide_rt: _peptide_rt_list) {
            p_peptide_rt->computeRetentionTimeReferences(_matching_mode);
        }
    }
}


void Quantificator::add_peptide_quanti_items(
    std::vector<const Peptide *> & isotope_peptides, std::vector<const Peptide *> & trace_peptide_list) {
    qDebug() << "Quantificator::add_peptide_quanti_items begin " << isotope_peptides.size();

    _p_peptide_list = new std::vector<const Peptide *>(isotope_peptides);

    /** \brief peptides charge to survey
     * */
    //std::map<const Peptide *, std::set<unsigned int>> map_peptides_charge_set;

    computeRetentionTimeReferences();

    //const XicExtractionMethodBase * extraction_method =
    //   _quanti_method->getXicExtractionMethod();

    pappso::PrecisionP extraction_precision = _quanti_method->getXicExtractionMethod()->getPappsoPrecision();
    qDebug() << "Quantificator::add_peptide_quanti_items extraction_precision " << extraction_precision->toString();
    QuantiItemBase * quanti_item;
    std::vector<const Peptide *>::const_iterator itpep;
    /// for each peptide (or isotope in the vector), we will create a peptide quanti item
    unsigned int  debug_i=0;
    for (itpep = isotope_peptides.begin(), debug_i=0; itpep != isotope_peptides.end();
            ++itpep, debug_i++) {
        mcqout() << "Quantificator::add_peptide_quanti_items for " << debug_i << " " << isotope_peptides.size() << endl;
        bool trace_peptide = false;
	const Peptide * original_peptide_p = (*itpep)->getOriginalPeptidePointer();
        if (std::find(trace_peptide_list.begin(), trace_peptide_list.end(), original_peptide_p) != trace_peptide_list.end()) {
            trace_peptide = true;
        }

        pappso::PeptideSp peptide_sp = (*itpep)->getPappsoPeptideSp();
        mcqout() << (*itpep)->getXmlId() << " " << peptide_sp.get()->toAbsoluteString() << endl;
        mcqout().flush();

        /// get the set of charges z the peptide is observed in the current group,
        /// and construct a quantification item for each peptide/z

        //std::set<unsigned int> * s_charges = (*itpep)->getCharges(
        //      *_group_to_quantify);
        std::vector<unsigned int> v_charges = (*itpep)->getPeptideRtSp()->getCharges();

        mcqout() << "charge OK " << endl;
        mcqout().flush();

        //if minimum natural isotope is set :
        // compute with natural isotope
        pappso::PeptideNaturalIsotopeList * _p_isotope_list = nullptr;
        if (_minimum_isotope_abundance > 0) {
            qDebug() << "if (_minimum_isotope_abundance > 0)";
            _p_isotope_list = new pappso::PeptideNaturalIsotopeList(peptide_sp);
            mcqout() << "PeptideNaturalIsotopeList OK " << endl;
            mcqout().flush();
        }

        // for each z do

        std::vector<unsigned int>::const_iterator itz;
        for (itz = v_charges.begin(); itz != v_charges.end(); ++itz) {
            if (_p_isotope_list != nullptr) {

                qDebug() << "Quantificator::add_peptide_quanti_items (_p_isotope_list != nullptr) *itz=" << *itz << " _p_isotope_list->size()=" << _p_isotope_list->size() << " pep xml id=" << (*itpep)->getXmlId() << " peptide=" << peptide_sp.get()->toAbsoluteString() ;
                std::vector<pappso::PeptideNaturalIsotopeAverageSp> natural_isotope_average_list = _p_isotope_list->getByIntensityRatio(*itz, extraction_precision, _minimum_isotope_abundance);
                mcqout() << "natural_isotope_average_list ok" << endl;
                mcqout().flush();
                qDebug() << "Quantificator::add_peptide_quanti_items natural_isotope_average_list";
                for (pappso::PeptideNaturalIsotopeAverageSp isotope_average_sp : natural_isotope_average_list) {
                    qDebug() << "average isotope " << isotope_average_sp.get()->getIntensityRatio() << " level=" <<  isotope_average_sp.get()->getIsotopeNumber() << " rank=" <<  isotope_average_sp.get()->getIsotopeRank() << endl;
                    quanti_item = new QuantiItemPeptideNaturalIsotope(trace_peptide, *itpep,isotope_average_sp);
                    _quanti_items->push_back(quanti_item);
                }

            }
            else {
                quanti_item = new QuantiItemPeptide(trace_peptide, *itpep,
                                                    *itz);
                _quanti_items->push_back(quanti_item);
            }
        }

        delete _p_isotope_list;
        _p_isotope_list = nullptr;
    }

    qDebug() << "Quantificator::add_peptide_quanti_items end";
}

void Quantificator::add_mz_quanti_items(const QStringList & mz_list) {
    QuantiItemBase * quanti_item;

    QStringList::const_iterator it;

    for (it = mz_list.begin(); it != mz_list.end(); ++it) {
        quanti_item = new QuantiItemBase(false,it->toDouble());
        _quanti_items->push_back(quanti_item);
    }
}

void Quantificator::add_mzrt_quanti_item(const QString & mz,
        const QString & rt) {

    QuantiItemBase * quanti_item = new QuantiItemMzRt(false,mz.toDouble(),
            rt.toDouble());
    _quanti_items->push_back(quanti_item);
}

void Quantificator::setMonitor(MonitorSpeedInterface * monitor) {
    _p_monitor_speed = monitor;
}


struct CmpQuantiItems {
    bool operator()(const QuantiItemBase * l1, const QuantiItemBase * l2) {
        const mcq_double mz1 = l1->getMz();
        const mcq_double mz2 = l2->getMz();
        return (mz1 < mz2);
    }
} QuantiItemsCmp;

void Quantificator::sortQuantiItems() {
    sort(_quanti_items->begin(), _quanti_items->end(), QuantiItemsCmp);
}



struct MapPrepareMsRun
{
    XicExtractionMethodBase * _p_xic_extraction_method;
    MapPrepareMsRun(XicExtractionMethodBase * p_xic_extraction_method) {
        _p_xic_extraction_method = p_xic_extraction_method;
    }

    MapPrepareMsRun(const MapPrepareMsRun& copy) {
        _p_xic_extraction_method = copy._p_xic_extraction_method;
    }

    void operator()(std::pair<QString, Msrun *> pair_msrun)
    {
        try {
            pair_msrun.second->prepareSpectraForQuantification(_p_xic_extraction_method);
        }
        catch (mcqError& errorException)
        {
            // QObject::tr("%1 CPUs used").arg(cpu_number)
            // mcqerr() << "Oops! an error occurred in MassChroQ. Dont Panic :" << endl;
            // mcqerr() << error.qwhat() << endl;
            qDebug() << " MapPrepareMsRun::operator " << errorException.qwhat();
            // windaube_exit();
            // return 1;
        }
    }
};

void Quantificator::quantify() {

    qDebug() << "Quantificator::quantify() begin";
    sortQuantiItems();

    QString group_id = _group_to_quantify->getXmlId();
    QString quanti_method_id = _quanti_method->getXmlId();
    const XicExtractionMethodBase * xic_extraction_method =
        _quanti_method->getXicExtractionMethod();
    McqXicType xic_type = _quanti_method->getXicType();
    const pappso::XicFilterInterface * _p_xic_filter = _quanti_method->getXicFilters();

    const pappso::XicDetectionInterface * detection_method = _quanti_method->getDetectionMethod();

    mcqout() << "MS run group '" << group_id << "': quantification method '"
             << quanti_method_id << "': quantification begin" << endl;
    _quanti_method->printInfos(mcqout());


    qDebug() << "Quantificator::quantify() writeQuantifyBegin";
    _p_monitor_speed->writeQuantifyBegin(this);
    //_p_monitor_speed->setCurrentGroup(_group_to_quantify);

    // launch quantification in group
    Msrun * current_msrun;
    msRunHashGroup::const_iterator itmsrun;
    unsigned int count_msruns(1);
    std::vector<QuantiItemBase *>::const_iterator itmz;

    if (_matching_mode == McqMatchingMode::realxic_or_mean) {

        count_msruns = 1;
        mcqout() << "_quanti_items->size() '" << _quanti_items->size() << endl;
        for (itmsrun = _group_to_quantify->begin();
                itmsrun != _group_to_quantify->end(); ++itmsrun, ++count_msruns) {
            current_msrun = itmsrun->second;

            mcqout() << "Preparing MS run " << current_msrun->getXmlId() << endl;
            current_msrun->prepareSpectraForQuantification(xic_extraction_method);

            mcqout() << "\tQuantifying (first pass) in MS run '" << (current_msrun->getXmlId())
                     << "' : " << count_msruns << "/" << _group_to_quantify->size()
                     << " in group '" << group_id << "'" << endl;

            MapQuantiRealXic mapQuanti (*_p_monitor_speed, current_msrun, xic_type,this,_p_xic_filter, detection_method);
            QFuture<void> res = QtConcurrent::map(_quanti_items->begin(), _quanti_items->end(), mapQuanti);
            res.waitForFinished();


            current_msrun->doneWithSpectra();

            mcqout() << "\t_p_peptide_list->size() '" << _p_peptide_list->size() << endl;
        }

        //compute retention time references
        for (auto && p_peptide_rt: _peptide_rt_list) {
            p_peptide_rt->computeRetentionTimeReferences(_matching_mode);
        }
    }
    count_msruns = 1;
    for (itmsrun = _group_to_quantify->begin();
            itmsrun != _group_to_quantify->end(); ++itmsrun, ++count_msruns) {
        current_msrun = itmsrun->second;
        mcqout() << "Preparing MS run " << current_msrun->getXmlId() << endl;
        current_msrun->prepareSpectraForQuantification(xic_extraction_method);


        mcqout() << "\tQuantifying in MS run '" << current_msrun->getXmlId()
                 << "' : " << count_msruns << "/" << _group_to_quantify->size()
                 << " in group '" << group_id << "'" << endl;

        MapQuanti mapQuanti (*_p_monitor_speed, current_msrun, xic_type,this,_p_xic_filter, detection_method);

        QFuture<void> res = QtConcurrent::map(_quanti_items->begin(), _quanti_items->end(), mapQuanti);
        res.waitForFinished();


        current_msrun->doneWithSpectra();

    }

    if (_matching_mode == McqMatchingMode::post_matching) {
        //compute new rt references
        mcqout() << "post matching : computing retention time references " << endl;
        QFuture<void> res = QtConcurrent::map(_peptide_rt_list.begin(), _peptide_rt_list.end(), [] (PeptideRt * p_peptide_rt) {
            p_peptide_rt->computePostMatchingTimeReferences();
        });
        res.waitForFinished();

        //rematch peaks stored in quanti items
        mcqout() << "post matching : trying to match peaks using new retention time references " << endl;
        MonitorSpeedInterface * p_monitor_speed = _p_monitor_speed;
        res = QtConcurrent::map(_quanti_items->begin(), _quanti_items->end(), [p_monitor_speed] (QuantiItemBase * p_quanti_item) {
            p_quanti_item->doPostMatching(*p_monitor_speed);
        });
        res.waitForFinished();
    }


    _p_monitor_speed->writeQuantifyEnd();

    qDebug() << "Quantificator::quantify() end";

}


void Quantificator::addMissedQuantiItem(const Msrun* _p_current_msrun, const QuantiItemBase* _p_current_search_item) {
    QMutexLocker locker(&_mutex);
    auto it_missed = _missed_quanti_item.insert(
                         std::pair<const Msrun*, std::vector<const QuantiItemBase*>> (_p_current_msrun, std::vector<const QuantiItemBase*>()));
    it_missed.first->second.push_back(_p_current_search_item);

}

bool Quantificator::isInMissedQuantiItem(const Msrun* _p_msrun, const QuantiItemBase* p_currentSearchItem) const {
    auto it = _missed_quanti_item.find(_p_msrun);
    if (it == _missed_quanti_item.end()) {

        qDebug() << "Quantificator::isInMissedQuantiItem false no msrun end ";
        return false;
    }
    else {
        auto itquanti = std::find (it->second.begin(), it->second.end(), p_currentSearchItem);
        if (itquanti == it->second.end()) {

            qDebug() << "Quantificator::isInMissedQuantiItem false no quanti item end ";
            return false;
        }
    }
    return true;
}

